/*
 * Copyright (C) 2019 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 */
package org.dripdronescanner.android.data;

import android.os.Build;

import androidx.annotation.NonNull;
import androidx.annotation.RequiresApi;
import androidx.lifecycle.MutableLiveData;

import org.dripdronescanner.android.Constants;

public class AircraftObject {
    final public MutableLiveData<Connection> connection = new MutableLiveData<>();
    final public MutableLiveData<Identification> identification = new MutableLiveData<>();
    final public MutableLiveData<LocationData> location = new MutableLiveData<>();
    final public MutableLiveData<AuthenticationData> authentication = new MutableLiveData<>();
    final public MutableLiveData<SelfIdData> selfid = new MutableLiveData<>();
    final public MutableLiveData<SystemData> system = new MutableLiveData<>();
    final public MutableLiveData<OperatorIdData> operatorid = new MutableLiveData<>();

    private final long macAddress;
    private byte[] pubkey;
    public AircraftObject(long macAddress) {
        this.macAddress = macAddress;
        this.pubkey = new byte[0];
    }
    public long getMacAddress() { return macAddress; }

    public Connection getConnection() { return connection.getValue(); }
    public Identification getIdentification() {
        return identification.getValue();
    }
    public LocationData getLocation() { return location.getValue(); }
    public AuthenticationData getAuthentication() { return authentication.getValue(); }
    public SelfIdData getSelfID() { return selfid.getValue(); }
    public SystemData getSystem() { return system.getValue(); }
    public OperatorIdData getOperatorID() { return operatorid.getValue(); }

    // Non-zero authentication data pages do not contain the following fields. Save them for displaying
    private int authPageCountSave;
    private int authLengthSave;
    private long authTimestampSave;

    // Multiple authentication messages are possible, each transmitting a part of the
    // authentication signature. Collect the data into authDataCombined.
    private byte[] authDataCombined = new byte[Constants.MAX_AUTH_DATA];

    @RequiresApi(api = Build.VERSION_CODES.O)
    public AuthenticationData combineAuthentication(AuthenticationData newData) {
        AuthenticationData currData = authentication.getValue();
        if (currData == null)
            currData = new AuthenticationData(newData.getAircraft());

        currData.setADCounter(newData.getADCounter());
        currData.setTimestamp(newData.getTimestamp());

        int offset = 0;
        int amount = Constants.MAX_AUTH_PAGE_ZERO_SIZE;
        if (newData.getAuthDataPage() == 0)  {
            authPageCountSave = newData.getAuthPageCount();
            authLengthSave = newData.getAuthLength();
            authTimestampSave = newData.getAuthTimestamp();
        } else {
            offset = Constants.MAX_AUTH_PAGE_ZERO_SIZE + (newData.getAuthDataPage() - 1) * Constants.MAX_AUTH_PAGE_NON_ZERO_SIZE;
            amount = Constants.MAX_AUTH_PAGE_NON_ZERO_SIZE;
        }
        for (int i = offset; i < offset + amount; i++)
            authDataCombined[i] = newData.getAuthData()[i];

        currData.setAuthType(newData.getAuthType());
        currData.setAuthPageCount(authPageCountSave);
        currData.setAuthLength(authLengthSave);
        currData.setAuthTimestamp(authTimestampSave);
        currData.setAuthData(authDataCombined);
        //Should always be authenticated but current validate function only works on drip
        if ((authPageCountSave - 1) == newData.getAuthDataPage()){
            // returns boolean but also sets values on currData therefore
            // result can for this instance be ignored
            boolean isValid = currData.validate(identification.getValue().getUasId(), this.pubkey);
            if(!isValid){
                currData.setServerIssues(this.pubkey.length == 0);
            }else{
                currData.setServerIssues(false);
            }
        }
        return currData;
    }

    public byte[] getPubkey() {
        if(this.pubkey == null) return new byte[0];
        return this.pubkey;
    }

    @Override @NonNull
    public String toString() {
        return "AircraftObject{" +
                "macAddress=" + macAddress +
                ", identification=" + identification +
                '}';
    }

    public void setPubKey(byte[] pubKey) {
        if(pubKey == null){
            this.pubkey = new byte[0];
        }else {
            this.pubkey = pubKey;
        }
    }
}
