/*
 * Copyright (C) 2019 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 */
package org.dripdronescanner.android.data;

import static org.dripdronescanner.android.Constants.ED25519_SIG_LENGTH;
import static org.dripdronescanner.android.Constants.MAX_DRIP_ID_BYTE_SIZE;

import android.os.Build;

import androidx.annotation.RequiresApi;

import org.bouncycastle.math.ec.rfc8032.Ed25519;
import org.dripdronescanner.android.Constants;

import java.sql.Timestamp;
import java.time.Instant;
import java.util.Arrays;
import java.util.Locale;

public class AuthenticationData extends MessageData {
    public static int TIMESTAMP_OFFSET = 1546300800;

    private AuthTypeEnum authType;
    private AircraftObject aircraft;
    private int authDataPage;
    private int authPageCount;
    private int authLength;
    private long authTimestamp;


    /**
     * reference to id type used to determine
     * how to validate auth message.
     */
    private int authIDType;
    private byte[] authData;
    private boolean isValidated;
    private boolean serverIssues;

    public AuthenticationData(AircraftObject ac) {
        super();
        aircraft = ac;
        authType = AuthTypeEnum.None;
        authDataPage = 0;
        authPageCount = 0;
        authLength = 0;
        authTimestamp = 0;
        authData = new byte[0];
        isValidated = false;
        serverIssues = false;
    }

    public enum AuthTypeEnum {
        None(0),
        UAS_ID_Signature(1),
        Operator_ID_Signature(2),
        Message_Set_Signature(3),
        Network_Remote_ID(4),
        Private_Use_0xA(0xA),
        Private_Use_0xB(0xB),
        Private_Use_0xC(0xC),
        Private_Use_0xD(0xD),
        Private_Use_0xE(0xE),
        Private_Use_0xF(0xF);

        AuthTypeEnum(int id) { this.id = id; }
        public final int id;
    }

    public AircraftObject getAircraft() {
        return aircraft;
    }

    public AuthTypeEnum getAuthType() { return authType; }
    void setAuthType(AuthTypeEnum authType) { this.authType = authType; }
    public void setAuthType(int authType) {
        switch(authType) {
            case 1: this.authType = AuthTypeEnum.UAS_ID_Signature; break;
            case 2: this.authType = AuthTypeEnum.Operator_ID_Signature; break;
            case 3: this.authType = AuthTypeEnum.Message_Set_Signature; break;
            case 4: this.authType = AuthTypeEnum.Network_Remote_ID; break;
            case 0xA: this.authType = AuthTypeEnum.Private_Use_0xA; break;
            case 0xB: this.authType = AuthTypeEnum.Private_Use_0xB; break;
            case 0xC: this.authType = AuthTypeEnum.Private_Use_0xC; break;
            case 0xD: this.authType = AuthTypeEnum.Private_Use_0xD; break;
            case 0xE: this.authType = AuthTypeEnum.Private_Use_0xE; break;
            case 0xF: this.authType = AuthTypeEnum.Private_Use_0xF; break;
            default: this.authType = AuthTypeEnum.None; break;
        }
    }

    int getAuthDataPage() { return authDataPage; }
    public void setAuthDataPage(int authDataPage) {
        if (authDataPage < 0)
            authDataPage = 0;
        if (authDataPage > (getMaxAuthPages()-1))
            authDataPage = getMaxAuthPages()-1;
        this.authDataPage = authDataPage;
    }

    private int getMaxAuthPages(){
        if(this.authIDType == Identification.IdTypeEnum.DRIP_ID.getValue())
            return Constants.MAX_DRIP_AUTH_DATA_PAGES;
        return Constants.MAX_AUTH_DATA_PAGES;

    }

    int getAuthPageCount() { return authPageCount; }

    public String getAuthPageCountAsString() {
        return String.format(Locale.US,"%d pages", authPageCount);
    }
    public void setAuthPageCount(int authPageCount) {
        if (authPageCount < 0)
            authPageCount = 0;
        if (authPageCount > getMaxAuthPages())
            authPageCount = getMaxAuthPages();
        this.authPageCount = authPageCount;
    }

    int getAuthLength() { return authLength; }
    public String getAuthLengthAsString() {
        return String.format(Locale.US,"%d bytes", authLength);
    }
    public void setAuthLength(int authLength) {
        if (authLength < 0)
            authLength = 0;
        if (authLength > getMaxAuthData())
            authLength = getMaxAuthData();
        this.authLength = authLength;
    }

    private int getMaxAuthData(){
        if(this.authIDType == Identification.IdTypeEnum.DRIP_ID.getValue())
            return Constants.MAX_DRIP_AUTH_DATA;
        return Constants.MAX_AUTH_DATA;
    }

    long getAuthTimestamp() { return authTimestamp; }
    public String getAuthTimestampAsString() {
        if (authTimestamp == 0)
            return "Unknown";
        Timestamp time = new Timestamp((1546300800L + authTimestamp) * 1000);
        return time.toString();
    }
    public void setAuthTimestamp(long authTimestamp) { this.authTimestamp = authTimestamp; }


    byte[] getAuthData() { return authData; }
    public void setAuthData(byte[] authData) { this.authData = authData; }
    public String getAuthenticationDataAsString() {

        StringBuilder sb = new StringBuilder();
        if(aircraft.identification.getValue().getIdType() == Identification.IdTypeEnum.DRIP_ID) {
            if (this.isValidated){
                sb.append("Valid");
            }else {
                if(this.serverIssues){
                    sb.append("Unknown due to server issues or failed manual entry");
                }else {
                    sb.append("Invalid");
                }
            }
        }else {
            for (int i = 0; i < authLength; i++) {
                sb.append(String.format("%02X ", authData[i]));
            }
        }
        return sb.toString();
    }

    public void setServerIssues(boolean issues){
        this.serverIssues = issues;
    }

    public boolean isServerIssues(){
        return this.serverIssues;
    }

    /**
     *  Validates the authentication message. Currently this method
     *  can only validate DRIP authentication messages.
     * @param id The id for the drone
     * @param publicKey The public key for the drone
     * @return true if the authentication is valid, false otherwise.
     */
    @RequiresApi(api = Build.VERSION_CODES.O)
    public boolean validate(final byte[] id, final byte[] publicKey) {
        if(publicKey.length != Ed25519.PUBLIC_KEY_SIZE) return false;
        int hhitOffset = ED25519_SIG_LENGTH;
        int timeStampOffset = ED25519_SIG_LENGTH+MAX_DRIP_ID_BYTE_SIZE;

        byte[] hhit = Arrays.copyOfRange(authData, hhitOffset, timeStampOffset);
        if(!Arrays.equals(hhit, id))
            return false;

        int expiration = astmDecodeInt(Arrays.copyOfRange(authData, timeStampOffset, timeStampOffset+4));
        if(expiration <= getASTMTime())
            return false;

        byte[] sig = Arrays.copyOfRange(authData, 0, 64);
        byte[] msg = Arrays.copyOfRange(authData, ED25519_SIG_LENGTH, this.authLength);
       boolean signatureValid = Ed25519.verify(
                sig, 0, publicKey, 0, msg, 0, this.authLength-ED25519_SIG_LENGTH);
       this.isValidated = signatureValid;
        return signatureValid;
    }



    public int astmDecodeInt(byte[] val) {
        /* Assuming byte order is LE */
        int newVal = 0;
        for (int i = 0; i < val.length; i++) {
            newVal |= (val[i] & 0xFF) << (i*8); // & 0xFF to get unsigned byte
        }
        return newVal;
    }

    @RequiresApi(api = Build.VERSION_CODES.O)
    public static long getASTMTime() {
        return Instant.now().getEpochSecond() - TIMESTAMP_OFFSET;
    }
}
