/*
 * Copyright (C) 2019 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 */
package org.dripdronescanner.android.data;

import static org.dripdronescanner.android.Constants.ED25519_CURVE_ID;
import static org.dripdronescanner.android.Constants.HIPv2_CONTEXT_ID;

import android.util.Log;

import org.bouncycastle.crypto.digests.CSHAKEDigest;
import org.bouncycastle.util.encoders.Hex;
import org.dripdronescanner.android.Constants;

import java.util.Arrays;

public class Identification extends MessageData {

    private UaTypeEnum uaType;
    private IdTypeEnum idType;
    private byte ogaId;
    private byte[] uasId;
    private byte[] hid;
    private AircraftObject aircraft;
    private boolean valid;
    private boolean isValidated;


    public Identification(AircraftObject aircraftObject) {
        super();
        uaType = UaTypeEnum.None;
        idType = IdTypeEnum.None;
        uasId = new byte[0];
        aircraft = aircraftObject;
        valid = false;
        isValidated = false;

    }



    public enum UaTypeEnum {
        None,
        Aeroplane,
        Helicopter_or_Multirotor,
        Gyroplane,
        Hybrid_Lift, // VTOL. Fixed wing aircraft that can take off vertically
        Ornithopter,
        Glider,
        Kite,
        Free_balloon,
        Captive_balloon,
        Airship,
        Free_fall_parachute, // Unpowered
        Rocket,
        Tethered_powered_aircraft,
        Ground_obstacle,
        Other,
    }

    public enum IdTypeEnum {
        None(0),//0
        Serial_Number(1),//1
        CAA_Registration_ID(2),//2
        UTM_Assigned_ID(3),//3
        DRIP_ID(4);//4

        int value;

        IdTypeEnum(int val){
            this.value = val;
        }

        public int getValue(){
            return this.value;
        }


        public static boolean contains(int idType){
            if(0<= idType && idType <= 4){
                return true;
            }
            return false;
        }
    }

    public UaTypeEnum getUaType() { return uaType; }
    public void setUaType(int uaType) {
        switch(uaType) {
            case 1: this.uaType = UaTypeEnum.Aeroplane; break;
            case 2: this.uaType = UaTypeEnum.Helicopter_or_Multirotor; break;
            case 3: this.uaType = UaTypeEnum.Gyroplane; break;
            case 4: this.uaType = UaTypeEnum.Hybrid_Lift; break;
            case 5: this.uaType = UaTypeEnum.Ornithopter; break;
            case 6: this.uaType = UaTypeEnum.Glider; break;
            case 7: this.uaType = UaTypeEnum.Kite; break;
            case 8: this.uaType = UaTypeEnum.Free_balloon; break;
            case 9: this.uaType = UaTypeEnum.Captive_balloon; break;
            case 10: this.uaType = UaTypeEnum.Airship; break;
            case 11: this.uaType = UaTypeEnum.Free_fall_parachute; break;
            case 12: this.uaType = UaTypeEnum.Rocket; break;
            case 13: this.uaType = UaTypeEnum.Tethered_powered_aircraft; break;
            case 14: this.uaType = UaTypeEnum.Ground_obstacle; break;
            case 15: this.uaType = UaTypeEnum.Other; break;
            default: this.uaType = UaTypeEnum.None; break;
        }
    }

    public IdTypeEnum getIdType() { return idType; }
    public void setIdType(int idType) {
        switch(idType) {
            case 1: this.idType = IdTypeEnum.Serial_Number; break;
            case 2: this.idType = IdTypeEnum.CAA_Registration_ID; break;
            case 3: this.idType = IdTypeEnum.UTM_Assigned_ID; break;
            case 4: this.idType = IdTypeEnum.DRIP_ID;break;
            default: this.idType = IdTypeEnum.None; break;
        }
    }

    public byte[] getUasId() { return uasId; }
    public String getUasIdAsString() {
        if(this.idType == IdTypeEnum.DRIP_ID){
            return Hex.toHexString(uasId).replaceFirst("0x", "");
        }
        return new String(uasId);
    }

    public void setUasId(byte[] uasId) {
        if (uasId.length <= Constants.MAX_ID_BYTE_SIZE) {
            this.uasId = uasId;
        }
        hid = Arrays.copyOfRange(uasId, 4, 8); // hid from byte 4
        ogaId = (byte) (uasId[3] & 0x0F); // oga_id in bits 4-7 in byte 3
    }

    public String getIDValid() {
        if(this.isValidated && this.valid) return "Valid";
        if(this.isValidated) return "Invalid";
        return "Unknown due to server issues or failed manual entry";
    }

    public boolean isValid() {
        return valid;
    }

    @Override
    public boolean equals(Object o) {
        if (this == o)
            return true;

        if (o == null || getClass() != o.getClass())
            return false;

        Identification that = (Identification) o;
        return uaType == that.uaType &&
               idType == that.idType &&
               Arrays.equals(uasId, that.uasId);
    }

    public boolean validateId(byte[] pub_key) {
        if(pub_key == null){
            pub_key = new byte[0];
        }
        if(ogaId == 0x5) {
            byte[] message = new byte[HIPv2_CONTEXT_ID.length + ED25519_CURVE_ID.length + aircraft.getPubkey().length];
            System.arraycopy(HIPv2_CONTEXT_ID, 0, message, 0, HIPv2_CONTEXT_ID.length);
            System.arraycopy(ED25519_CURVE_ID, 0, message, HIPv2_CONTEXT_ID.length, ED25519_CURVE_ID.length);
            System.arraycopy(pub_key, 0, message, HIPv2_CONTEXT_ID.length + ED25519_CURVE_ID.length, aircraft.getPubkey().length);

            CSHAKEDigest dgst = new CSHAKEDigest(128, null, HIPv2_CONTEXT_ID);
            byte[] output = new byte[8];
            dgst.update(message, 0, message.length);
            dgst.doOutput(output, 0, 8);

            boolean idValidated = Arrays.equals(output, Arrays.copyOfRange(uasId, 8, 16));
            //setIdValidationStatus(idValidated ? "Valid" : "Invalid");
            this.isValidated = pub_key.length != 0;
            this.valid = idValidated;
            return idValidated;
        } else {
            this.isValidated = false;
            this.valid = false;
            Log.w("DRIP_ID", "Oga id not recognized during verification of id.");
            return false;
        }
    }
}
