package org.dripdronescanner.android.network;

import android.app.Activity;
import android.content.Context;
import android.content.SharedPreferences;
import android.util.JsonReader;
import android.util.Log;

import org.dripdronescanner.android.app.DebugActivity;
import org.dripdronescanner.android.log.LogWriter;
import org.json.JSONException;
import org.json.JSONObject;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.Callable;

import javax.net.ssl.HttpsURLConnection;

public class CertificateServerAPI implements Callable {
        static final Map<String,String> idToKeys = new HashMap<>();
    public CertificateServerAPI() {
    }


    /**
     * Fetches the public key for a given drip id. If key has already been fetched then
     * it will simply be returned directly. Note that this is dependant on the lifecycle for this
     * class. When the class is terminated any memory of keys will be forgotten.
     * @param activity The activity that was running when this is called.
     * @param basicId The ID object that is used to give the drip id.
     * @param logger Potential logger object, can be null.
     * @return A JSONObject in format {id1:{"drip_id":id1,"pub_key":pub_key1},id2:{...}...}
     * or null.
     */
    public static JSONObject getKey(Activity activity,OpenDroneIdParser.BasicId basicId,LogWriter logger)  {
        if(idToKeys.containsKey(basicId.toDRIP())){
            try {
                return new JSONObject().put(basicId.toDRIP(), idToKeys.get(basicId.toDRIP()));
            } catch (JSONException e) {
                System.out.println("Unable to parse json data while reading locally: " +e);
                if(logger != null) {
                    logger.logText("Unable to parse json data while reading locally: \" +e", null);
                }
                return null;
            }
        }
        if(logger == null){
            try {
                logger = new LogWriter(((DebugActivity) activity).getLoggerFileDir("certificateServerCall"));
            }catch(IOException e){
                System.out.println("Unable to create a logger");
                Log.i("CertificateServerCall","Unable to create a logger, might cause missed logs");
            }
        }
        SharedPreferences sp = activity.getSharedPreferences("ONLY_PREF",Context.MODE_PRIVATE);
        String base_url = sp.getString("url","https://192.168.0.1:8080");
        if (base_url.endsWith("/")) {
            base_url = base_url.substring(0,base_url.length()-2);
        }
        String server_url = base_url +"/get/key";
        JSONObject result = null;
        try {
            URL url = new URL(server_url);
            HttpsURLConnection urlConnection = (HttpsURLConnection) url.openConnection();
            try {
                urlConnection.setRequestMethod("POST");
                urlConnection.setRequestProperty("Content-Type","application/json;charset=UTF-8");
                urlConnection.setRequestProperty("Accept","application/json");
                urlConnection.setDoOutput(true);
                urlConnection.setHostnameVerifier((s, sslSession) -> {
                    //Allows for local lan connections only.
                    return s.startsWith("192.168") ||
                            s.startsWith("localhost") ||
                            s.startsWith("127.0.0.1") ||
                            s.startsWith("zebraland.ida.liu.se");
                });
                urlConnection.setChunkedStreamingMode(0);
                urlConnection.connect();
                OutputStream out = new BufferedOutputStream(urlConnection.getOutputStream());

                String data = "{\"drip_ids\": [\"" + basicId.toDRIP() + "\"]}";
                System.out.println("Sending data to certificate server:" + data);
                out.write(data.getBytes(StandardCharsets.UTF_8));
                out.flush();
                out.close();
                InputStream in = new BufferedInputStream(urlConnection.getInputStream());
                JsonReader reader = new JsonReader(new InputStreamReader(in, "UTF-8"));
                result = readJson(reader);

            } catch (IOException e) {
                e.printStackTrace();
                System.out.println("Something went wrong in writing to or receiving from the server:" + url.toString());
                if(logger != null) {
                    String [] data = new String[2];
                    data[0] = e.toString();
                    data[1] = url.toString();
                    logger.logText("Something went wrong in writing to or receiving from the server",data);
                }


            } catch (JSONException e) {
                System.out.println("Unable to parse json data: " + e);
                if(logger != null) {
                    logger.logText("Unable to parse json data: " + e, null);
                }

            } finally {
                urlConnection.disconnect();
            }

        } catch (IOException e) {
            System.out.println("Unable to create connection, no internet or server is not responding:" + e);
            if(logger != null) {
                logger.logText("Unable to create connection, no internet or server is not responding: " +e, null);
            }

            return null;
        }
        return result;
    }

    private static JSONObject readJson(JsonReader reader) throws IOException, JSONException {
        reader.beginArray();
        List<JSONObject> objects = new ArrayList<>();
        while(reader.hasNext()){
           objects.add(readObject(reader));
        }
        reader.endArray();
        JSONObject result = new JSONObject();
        for(int i = 0; i < objects.size(); i++){
            result.put(objects.get(i).getString("drip_id"), objects.get(i).getString("pub_key"));
        }
        return result;
    }

    private static JSONObject readObject(JsonReader reader) throws IOException, JSONException {
        String id = null;
        String pub_key = null;
        reader.beginObject();
        while (reader.hasNext()) {
            String name = reader.nextName();
            if (name.equals("drip_id")) {
                id = reader.nextString();
            } else if (name.equals("pub_key")) {
                pub_key = reader.nextString();
            } else {
                reader.skipValue();
            }
        }
        reader.endObject();
        idToKeys.put(id,pub_key);
        return new JSONObject().put("drip_id",id).put("pub_key",pub_key);

    }

    @Override
    public Object call() throws Exception {
        return null;
    }
}
