/*
 * Copyright (C) 2019 Intel Corporation
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 */
package org.dripdronescanner.android.network;

import android.app.Activity;
import android.util.Log;
import android.widget.Toast;

import androidx.annotation.NonNull;

import org.bouncycastle.util.encoders.Hex;
import org.dripdronescanner.android.Constants;
import org.dripdronescanner.android.data.Identification;
import org.dripdronescanner.android.log.LogMessageEntry;
import org.json.JSONException;
import org.json.JSONObject;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Locale;
import java.util.concurrent.CompletableFuture;


public class OpenDroneIdParser {
    private static final String TAG = "OpenDroneIdParser";
    private static final String DELIM = Constants.DELIM;

    public enum Type {
        BASIC_ID(0),
        LOCATION(1),
        AUTH(2),
        SELFID(3),
        SYSTEM(4),
        OPERATOR_ID(5),
        MESSAGE_PACK(0xF);

        Type(int id) { this.id = id; }
        public final int id;

        public static Type fromId(int id) {
            if (id == BASIC_ID.id) {
                return BASIC_ID;
            } else if (id == LOCATION.id) {
                return LOCATION;
            } else if (id == AUTH.id) {
                return AUTH;
            } else if (id == SELFID.id) {
                return SELFID;
            } else if (id == SYSTEM.id) {
                return SYSTEM;
            } else if (id == OPERATOR_ID.id) {
                return OPERATOR_ID;
            } else if (id == MESSAGE_PACK.id) {
                return MESSAGE_PACK;
            } else {
                return null;
            }
        }
    }

    public static class Header {
        public Type type;
        public int version;

        @Override @NonNull
        public String toString() {
            return "Header{" +
                    "type=" + type +
                    ", version=" + version +
                    '}';
        }
    }

    public interface Payload {
        String toCsvString();
    }

    private static final double LAT_LONG_MULTIPLIER = 1e-7;
    private static final double SPEED_VERTICAL_MULTIPLIER = 0.5;

    public static class BasicId implements Payload {
        int idType;
        int uaType;
        byte[] uasId;
        byte[] pubKey;

        public BasicId(int idType, int uaType,byte[] idmsg) {
            this.idType = idType;
            this.uaType = uaType;
            this.uasId = idmsg;
        }

        public BasicId(int idlength) {
            this.uasId = new byte[idlength];
        }

        public static String csvHeader() {
            return "idType" + DELIM
                    + "uaType" + DELIM
                    + "uasId" + DELIM;
        }

        public String toDRIP(){
            return Hex.toHexString(this.uasId).replaceFirst("0x", "");
        }

        public void setPubKey(byte[] pub_key) {
            if (pub_key == null){
                this.pubKey = new byte[0];
            }else {
                this.pubKey = pub_key;
            }
        }

        @Override
        public String toCsvString() {
            return idType + DELIM
                    + uaType + DELIM
                    + new String(uasId) + DELIM;
        }

        @Override @NonNull
        public String toString() {
            return "BasicId{" +
                    "idType=" + idType +
                    ", uaType=" + uaType +
                    ", uasId='" + Arrays.toString(uasId) + '\'' +
                    '}';
        }
    }

    public static class Location implements Payload {
        int status;
        int heightType;
        int EWDirection;
        int speedMult;
        int Direction;
        int speedHori;
        int speedVert;
        int droneLat;
        int droneLon;
        int altitudePressure;
        int altitudeGeodetic;
        int height;
        int horizontalAccuracy;
        int verticalAccuracy;
        int baroAccuracy;
        int speedAccuracy;
        int timestamp;
        int timeAccuracy;
        float distance;

        static double calcSpeed(int value, int mult) {
            if (mult == 0)
                return value * 0.25;
            else
                return (value * 0.75) + (255 * 0.25);
        }

        static double calcDirection(int value, int EW) {
            if (EW == 0)
                return value;
            else
                return value + 180;
        }

        double getDirection() { return calcDirection(Direction, EWDirection); }
        double getSpeedHori() { return calcSpeed(speedHori, speedMult); }

        double getSpeedVert() { return SPEED_VERTICAL_MULTIPLIER * speedVert; }

        double getLatitude() {
            return LAT_LONG_MULTIPLIER * droneLat;
        }
        double getLongitude() {
            return LAT_LONG_MULTIPLIER * droneLon;
        }

        static double calcAltitude(int value) { return (double) value / 2 - 1000; }
        double getAltitudePressure() { return calcAltitude(altitudePressure); }
        double getAltitudeGeodetic() { return calcAltitude(altitudeGeodetic); }
        double getHeight() { return calcAltitude(height); }

        double getTimeAccuracy() { return timeAccuracy * 0.1; }

        public static String csvHeader() {
            return "status" + DELIM
                    + "heightType" + DELIM
                    + "EWDirection" + DELIM
                    + "speedMult" + DELIM
                    + "direction" + DELIM
                    + "speedHori" + DELIM
                    + "speedVert" + DELIM
                    + "droneLat" + DELIM
                    + "droneLon" + DELIM
                    + "altitudePressure" + DELIM
                    + "altitudeGeodetic" + DELIM
                    + "height" + DELIM
                    + "horizontalAccuracy" + DELIM
                    + "verticalAccuracy" + DELIM
                    + "baroAccuracy" + DELIM
                    + "speedAccuracy" + DELIM
                    + "timestamp" + DELIM
                    + "timeAccuracy" + DELIM
                    + "distance" + DELIM;
        }

        @Override
        public String toCsvString() {
            return status + DELIM
                    + heightType + DELIM
                    + EWDirection + DELIM
                    + speedMult + DELIM
                    + Direction + DELIM
                    + speedHori + DELIM
                    + speedVert + DELIM
                    + droneLat + DELIM
                    + droneLon + DELIM
                    + altitudePressure + DELIM
                    + altitudeGeodetic + DELIM
                    + height + DELIM
                    + horizontalAccuracy + DELIM
                    + verticalAccuracy + DELIM
                    + baroAccuracy + DELIM
                    + speedAccuracy + DELIM
                    + timestamp + DELIM
                    + timeAccuracy + DELIM
                    + distance + DELIM;
        }
        
        @Override @NonNull
        public String toString() {
            return "Location{" +
                    "status=" + status +
                    ", heightType=" + heightType +
                    ", EWDirection=" + EWDirection +
                    ", speedMult=" + speedMult +
                    ", direction=" + Direction +
                    ", speedHori=" + speedHori +
                    ", speedVert=" + speedVert +
                    ", droneLat=" + droneLat +
                    ", droneLon=" + droneLon +
                    ", altitudePressure=" + altitudePressure +
                    ", altitudeGeodetic=" + altitudeGeodetic +
                    ", height=" + height +
                    ", horizontalAccuracy=" + horizontalAccuracy +
                    ", verticalAccuracy=" + verticalAccuracy +
                    ", baroAccuracy=" + baroAccuracy +
                    ", speedAccuracy=" + speedAccuracy +
                    ", timestamp=" + timestamp +
                    ", timeAccuracy=" + timeAccuracy +
                    ", distance=" + distance +
                    '}';
        }
    }

    public static class Authentication implements Payload {
        int authType;
        int authDataPage;
        int authPageCount;
        int authLength;
        int authTimestamp;
        byte[] authData = new byte[Constants.MAX_AUTH_DATA];

        public int getAuthDataPage() { return authDataPage; }

        public static String csvHeader() {
            return "authType" + DELIM
                    + "authDataPage" + DELIM
                    + "authPageCount" + DELIM
                    + "authLength" + DELIM
                    + "authTimestamp" + DELIM
                    + "authData" + DELIM;
        }

        private String authDataToString() {
            StringBuilder sb = new StringBuilder();
            for (byte authDatum : authData) {
                sb.append(String.format("%02X ", authDatum));
            }
            return sb.toString();
        }

        @Override
        public String toCsvString() {
            return authType + DELIM
                    + authDataPage + DELIM
                    + authPageCount + DELIM
                    + authLength + DELIM
                    + authTimestamp + DELIM
                    + authDataToString() + DELIM;
        }

        @Override @NonNull
        public String toString() {
            return "Authentication{" +
                    "authType=" + authType +
                    ", authDataPage=" + authDataPage +
                    ", authPageCount=" + authPageCount +
                    ", authLength=" + authLength +
                    ", authTimestamp=" + authTimestamp +
                    ", authData='" + Arrays.toString(authData) + '\'' +
                    '}';
        }
    }

    public static class SelfID implements Payload {
        int descriptionType;
        byte[] operationDescription = new byte[Constants.MAX_STRING_BYTE_SIZE];

        public static String csvHeader() {
            return "descriptionType" + DELIM
                    + "operationDescription" + DELIM;
        }

        @Override
        public String toCsvString() {
            return descriptionType + DELIM
                    + new String(operationDescription) + DELIM;
        }

        @Override @NonNull
        public String toString() {
            return "SelfID{" +
                    "descriptionType=" + descriptionType +
                    ", operationDescription='" + Arrays.toString(operationDescription) + '\'' +
                    '}';
        }
    }

    public static class SystemMsg implements Payload {
        int operatorLocationType;
        int classificationType;
        int operatorLatitude;
        int operatorLongitude;
        int areaCount;
        int areaRadius;
        int areaCeiling;
        int areaFloor;
        int category;
        int classValue;

        double getLatitude() {
            return LAT_LONG_MULTIPLIER * operatorLatitude;
        }
        double getLongitude() {
            return LAT_LONG_MULTIPLIER * operatorLongitude;
        }

        int getAreaRadius() { return areaRadius * 10; }
        static double calcAltitude(int value) { return (double) value / 2 - 1000; }
        double getAreaCeiling() { return calcAltitude(areaCeiling); }
        double getAreaFloor() { return calcAltitude(areaFloor); }

        public static String csvHeader() {
            return "operatorLocationType" + DELIM
                    + "classificationType" + DELIM
                    + "operatorLatitude" + DELIM
                    + "operatorLongitude" + DELIM
                    + "areaCount" + DELIM
                    + "areaRadius" + DELIM
                    + "areaCeiling" + DELIM
                    + "areaFloor" + DELIM
                    + "category" + DELIM
                    + "classValue" + DELIM;
        }

        @Override
        public String toCsvString() {
            return operatorLocationType + DELIM
                    + classificationType + DELIM
                    + operatorLatitude + DELIM
                    + operatorLongitude + DELIM
                    + areaCount + DELIM
                    + areaRadius + DELIM
                    + areaCeiling + DELIM
                    + areaFloor + DELIM
                    + category + DELIM
                    + classValue + DELIM;
        }

        @Override @NonNull
        public String toString() {
            return "PilotLocation{" +
                    "operatorLocationType=" + operatorLocationType +
                    ", classificationType=" + classificationType +
                    ", operatorLatitude=" + operatorLatitude +
                    ", operatorLongitude=" + operatorLongitude +
                    ", areaCount=" + areaCount +
                    ", areaRadius=" + areaRadius +
                    ", areaCeiling=" + areaCeiling +
                    ", areaFloor=" + areaFloor +
                    ", category=" + category +
                    ", class=" + classValue +
                    '}';
        }
    }

    public static class OperatorID implements Payload {
        int operatorIdType;
        byte[] operatorId = new byte[Constants.MAX_ID_BYTE_SIZE];

        public static String csvHeader() {
            return "operatorIdType" + DELIM
                    + "operatorId" + DELIM;
        }

        @Override
        public String toCsvString() {
            return operatorIdType + DELIM
                    + new String(operatorId) + DELIM;
        }
        @Override @NonNull
        public String toString() {
            return "OperatorID{" +
                    "operatorIdType=" + operatorIdType +
                    ", operatorId='" + Arrays.toString(operatorId) + '\'' +
                    '}';
        }
    }

    public static class MessagePack implements Payload {
        int messageSize;
        int messagesInPack;
        byte[] messages = new byte[Constants.MAX_MESSAGE_PACK_SIZE];

        @Override @NonNull
        public String toString() {
            return "MessagePack{" +
                    "messageSize=" + messageSize +
                    ", messagesInPack=" + messagesInPack +
                    ", messages='" + Arrays.toString(messages) + '\'' +
                    '}';
        }

        @Override public String toCsvString() { return null; }
    }

    public static class Message<T extends Payload> implements Comparable<Message<T>> {
        final int adCounter;
        final long timestamp;
        public final Header header;
        public final T payload;

        Message(Header header, T payload, long timestamp, int adCounter) {
            this.adCounter = adCounter;
            this.header = header;
            this.payload = payload;
            this.timestamp = timestamp;
        }

        @Override
        public int compareTo(@NonNull Message o)
        {
            if (this.header.type == Type.AUTH && o.header.type == Type.AUTH ) {
                Authentication authThis = (OpenDroneIdParser.Authentication) this.payload;
                Authentication authO = (OpenDroneIdParser.Authentication) o.payload;
                return (authThis.authDataPage - authO.authDataPage);
            } else {
                return this.header.type.compareTo(o.header.type);
            }
        }
    }

    static Message<Payload> parseAdvertisingData(byte[] payload, int offset, long timestamp,
                                                 LogMessageEntry logMessageEntry,
                                                 android.location.Location receiverLocation,Activity activity) {
        if (offset <= 0 || payload.length < offset + Constants.MAX_MESSAGE_SIZE)
            return null;

        int adCounter = payload[offset - 1] & 0xFF;
        return parseMessage(payload, offset, timestamp, logMessageEntry, receiverLocation, adCounter,activity);
    }

    static Message<Payload> parseMessage(byte[] payload, int offset, long timestamp,
                                         LogMessageEntry logMessageEntry,
                                         android.location.Location receiverLocation, int adCounter, Activity activity) {
        if (payload.length < offset + Constants.MAX_MESSAGE_SIZE)
            return null;

        ByteBuffer byteBuffer = ByteBuffer.wrap(payload, offset, Constants.MAX_MESSAGE_SIZE);
        byteBuffer.order(ByteOrder.LITTLE_ENDIAN);

        Header header = new Header();
        int b = byteBuffer.get() & 0xFF;
        int type = (b & 0xF0) >> 4;
        header.type = Type.fromId(type);
        if (header.type == null) {
            Log.e(TAG, "Header type unknown");
            return null;
        }
        header.version = b & 0x0F;

        Payload payloadObj = null;

        switch (header.type) {
            case BASIC_ID:
                payloadObj = parseBasicId(byteBuffer,activity);
                break;
            case LOCATION:
                payloadObj = parseLocation(byteBuffer, receiverLocation);
                break;
            case AUTH:
                payloadObj = parseAuthentication(byteBuffer);
                break;
            case SELFID:
                payloadObj = parseSelfID(byteBuffer);
                break;
            case SYSTEM:
                payloadObj = parseSystem(byteBuffer);
                break;
            case OPERATOR_ID:
                payloadObj = parseOperatorID(byteBuffer);
                break;
            case MESSAGE_PACK:
                payloadObj = parseMessagePack(payload, offset);
                break;
            default:
                Log.w(TAG, "Received unhandled message type: id=" + type);

        }
        Message<Payload> message = new Message<>(header, payloadObj, timestamp, adCounter);
        if (header.type != Type.MESSAGE_PACK)
            logMessageEntry.add(message);
        return message;
    }


    private static BasicId parseBasicId(ByteBuffer byteBuffer, Activity activity) {
        BasicId basicId;

        int type = byteBuffer.get();
        int idType = (type & 0xF0) >> 4;
        int uaType  = type & 0x0F;
        if (idType == Identification.IdTypeEnum.DRIP_ID.getValue()) {
            basicId = new BasicId(Constants.MAX_DRIP_ID_BYTE_SIZE);
            basicId.idType = idType;
            basicId.uaType = uaType;
            byte[] idArr = byteBuffer.array();
            basicId.uasId = Arrays.copyOfRange(idArr, 8, 8+Constants.MAX_DRIP_ID_BYTE_SIZE);
            //fetch url from id message
            CompletableFuture<JSONObject> uaDataFuture = CompletableFuture
                    .supplyAsync(() -> CertificateServerAPI.getKey(activity,basicId,null));
            uaDataFuture.handle((keys,error) -> {
                if(keys == null || error != null){
                    System.out.println("No key retrieved due to: " + error);

                    Toast.makeText(activity.getApplicationContext(),
                            "Unable to retrieve key for id: " + basicId.toDRIP(),
                            Toast.LENGTH_LONG).show();
                }else{
                    try {
                        String pub_key = ((String)((JSONObject)keys).get(basicId.toDRIP())).toLowerCase(Locale.ROOT);
                        basicId.setPubKey(Hex.decode(pub_key.getBytes(StandardCharsets.UTF_8)));
                    } catch (JSONException e) {
                        System.out.println("Unable to parse json data after getting it from api: " + e);
                        Toast.makeText(activity.getApplicationContext(),
                                "Unable to parse data correctly",
                                Toast.LENGTH_LONG).show();
                    }
                }
                return basicId;
            });
        }else {
            basicId = new BasicId(Constants.MAX_ID_BYTE_SIZE);
            basicId.idType = idType;
            basicId.uaType = uaType;
            byteBuffer.get(basicId.uasId, 0, Constants.MAX_ID_BYTE_SIZE);
        }
        return basicId;
    }

    private static Location parseLocation(ByteBuffer byteBuffer,
                                          android.location.Location receiverLocation) {
        Location location = new Location();

        int b = byteBuffer.get();

        location.status = (b & 0xF0) >> 4;
        location.heightType = (b & 0x04) >> 2;
        location.EWDirection = (b & 0x02) >> 1;
        location.speedMult = b & 0x01;

        location.Direction = byteBuffer.get() & 0xFF;
        location.speedHori = byteBuffer.get() & 0xFF;
        location.speedVert = byteBuffer.get();

        location.droneLat = byteBuffer.getInt();
        location.droneLon = byteBuffer.getInt();

        location.altitudePressure = byteBuffer.getShort();
        location.altitudeGeodetic = byteBuffer.getShort();
        location.height = byteBuffer.getShort();

        int horiVertAccuracy = byteBuffer.get();
        location.horizontalAccuracy = horiVertAccuracy & 0x0F;
        location.verticalAccuracy = (horiVertAccuracy & 0xF0) >> 4;
        int speedBaroAccuracy = byteBuffer.get();
        location.baroAccuracy = (speedBaroAccuracy & 0xF0) >> 4;
        location.speedAccuracy = speedBaroAccuracy & 0x0F;
        location.timestamp = byteBuffer.getShort() & 0xFFFF;
        location.timeAccuracy = byteBuffer.get() & 0x0F;

        // Use an older retrieved receiver location to calculate the distance to the drone
        if (location.droneLat != 0 && location.droneLon != 0) {
            android.location.Location droneLoc = new android.location.Location("");
            droneLoc.setLatitude(location.getLatitude());
            droneLoc.setLongitude(location.getLongitude());
            if (receiverLocation != null)
                location.distance = receiverLocation.distanceTo(droneLoc);
        }

        return location;
    }

    private static Authentication parseAuthentication(ByteBuffer byteBuffer) {
        Authentication authentication = new Authentication();

        int type = byteBuffer.get();
        authentication.authType = (type & 0xF0) >> 4;
        authentication.authDataPage = type & 0x0F;

        int offset = 0;
        int amount = Constants.MAX_AUTH_PAGE_ZERO_SIZE;
        if (authentication.authDataPage == 0) {
            authentication.authPageCount = byteBuffer.get();
            authentication.authLength = byteBuffer.get();
            authentication.authTimestamp = byteBuffer.getInt();
        } else {
            offset = Constants.MAX_AUTH_PAGE_ZERO_SIZE + (authentication.authDataPage - 1) * Constants.MAX_AUTH_PAGE_NON_ZERO_SIZE;
            amount = Constants.MAX_AUTH_PAGE_NON_ZERO_SIZE;
        }
        if (authentication.authDataPage >= 0 && authentication.authDataPage < Constants.MAX_AUTH_DATA_PAGES)
            for (int i = offset; i < offset + amount; i++)
                authentication.authData[i] = byteBuffer.get();
        return authentication;
    }

    private static SelfID parseSelfID(ByteBuffer byteBuffer) {
        SelfID selfID = new SelfID();
        selfID.descriptionType = byteBuffer.get();
        byteBuffer.get(selfID.operationDescription, 0, Constants.MAX_STRING_BYTE_SIZE);
        return selfID;
    }

    private static SystemMsg parseSystem(ByteBuffer byteBuffer) {
        SystemMsg s = new SystemMsg();

        int b = byteBuffer.get();
        s.operatorLocationType = b & 0x03;
        s.classificationType = (b & 0x1C) >> 2;
        s.operatorLatitude = byteBuffer.getInt();
        s.operatorLongitude = byteBuffer.getInt();
        s.areaCount = byteBuffer.getShort() & 0xFFFF;
        s.areaRadius = byteBuffer.get() & 0xFF;
        s.areaCeiling = byteBuffer.getShort();
        s.areaFloor = byteBuffer.getShort();
        b = byteBuffer.get();
        s.category = (b & 0xF0) >> 4;
        s.classValue = b & 0x0F;
        return s;
    }

    private static OperatorID parseOperatorID(ByteBuffer byteBuffer) {
        OperatorID operatorID = new OperatorID();
        operatorID.operatorIdType = byteBuffer.get();
        byteBuffer.get(operatorID.operatorId, 0, Constants.MAX_ID_BYTE_SIZE);
        return operatorID;
    }


    private static MessagePack parseMessagePack(byte[] payload, int offset) {
        ByteBuffer byteBuffer = ByteBuffer.wrap(payload, offset + 1, 2);
        byteBuffer.order(ByteOrder.LITTLE_ENDIAN);

        MessagePack messagePack = new MessagePack();
        messagePack.messageSize = byteBuffer.get();
        messagePack.messagesInPack = byteBuffer.get();

        if (messagePack.messageSize != Constants.MAX_MESSAGE_SIZE ||
            messagePack.messagesInPack <= 0 ||
            messagePack.messagesInPack > Constants.MAX_MESSAGES_IN_PACK ||
            payload.length < offset + 1 + 2 + messagePack.messageSize*messagePack.messagesInPack)
            return null;

        // Now that we know how much data is in the message, re-wrap and extract the data
        byteBuffer = ByteBuffer.wrap(payload, offset + 1 + 2, messagePack.messageSize*messagePack.messagesInPack);
        byteBuffer.get(messagePack.messages, 0, messagePack.messageSize * messagePack.messagesInPack);
        return messagePack;
    }
}
